from pyfileconf import Selector
from tests.input_files.amodule import SecondExampleClass
from tests.input_files.mypackage.cmodule import ExampleClass
from tests.test_pipeline_manager.base import PipelineManagerTestBase, CLASS_CONFIG_DICT_LIST, SAME_CLASS_CONFIG_DICT_LIST, \
    DIFFERENT_CLASS_CONFIG_DICT_LIST


class TestPipelineManagerRun(PipelineManagerTestBase):

    def test_run_function(self):
        self.write_a_function_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.stuff.a_function
        result = pipeline_manager.run(iv)
        assert result == (None, None)

    def test_run_function_multiple_pms(self):
        self.write_a_function_to_pipeline_dict_file()
        self.write_a_function_to_pipeline_dict_file(file_path=self.second_pipeline_dict_path)
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        pipeline_manager2 = self.create_pm(
            folder=self.second_pm_folder,
            name=self.second_test_name,
        )
        pipeline_manager2.load()
        sel = Selector()

        # Assert pipeline manager 1 run
        iv = sel.test_pipeline_manager.stuff.a_function
        result = pipeline_manager.run(iv)
        assert result == (None, None)

        # Assert pipeline manager 2 run
        iv = sel.test_pipeline_manager2.stuff.a_function
        result = pipeline_manager.run(iv)
        assert result == (None, None)

    def test_create_class(self):
        self.write_example_class_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        sel = Selector()
        ec = sel.test_pipeline_manager.stuff.ExampleClass()
        assert ec == ExampleClass(None)

    def test_create_class_multiple_pms(self):
        self.write_example_class_to_pipeline_dict_file()
        self.write_example_class_to_pipeline_dict_file(file_path=self.second_pipeline_dict_path)
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        pipeline_manager2 = self.create_pm(
            folder=self.second_pm_folder,
            name=self.second_test_name,
        )
        pipeline_manager2.load()
        sel = Selector()

        # Assert pipeline manager 1 create
        ec = sel.test_pipeline_manager.stuff.ExampleClass()
        assert ec == ExampleClass(None)

        # Assert pipeline manager 2 create
        ec = sel.test_pipeline_manager2.stuff.ExampleClass()
        assert ec == ExampleClass(None)

    def test_create_from_specific_class_dict(self):
        self.write_example_class_dict_to_file()
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST
        )
        pipeline_manager.load()
        sel = Selector()
        ec = sel.test_pipeline_manager.example_class.stuff.data
        expect_ec = ExampleClass(None, name='data')
        assert ec.name == expect_ec.name
        assert ec.a == expect_ec.a

    def test_create_from_specific_class_dict_multiple_pms(self):
        self.write_example_class_dict_to_file()
        self.write_example_class_dict_to_file(pm_index=1)
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST
        )
        pipeline_manager.load()
        pipeline_manager2 = self.create_pm(
            folder=self.second_pm_folder,
            name=self.second_test_name,
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST,
        )
        pipeline_manager2.load()
        sel = Selector()

        # Assert pipeline manager 1 create
        ec = sel.test_pipeline_manager.example_class.stuff.data
        expect_ec = ExampleClass(None, name='data')
        assert ec.name == expect_ec.name
        assert ec.a == expect_ec.a

        # Assert pipeline manager 2 create
        ec = sel.test_pipeline_manager2.example_class.stuff.data
        expect_ec = ExampleClass(None, name='data')
        assert ec.name == expect_ec.name
        assert ec.a == expect_ec.a

    def test_create_from_multiple_specific_class_dicts_same_class(self):
        self.write_example_class_dict_to_file()  # example_class
        self.write_example_class_dict_to_file(1)  # example_class2
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=SAME_CLASS_CONFIG_DICT_LIST
        )
        pipeline_manager.load()
        sel = Selector()
        ec = sel.test_pipeline_manager.example_class.stuff.data
        ec2 = sel.test_pipeline_manager.example_class2.stuff.data
        expect_ec = ExampleClass(None, name='data')
        assert ec.name == ec2.name == expect_ec.name
        assert ec.a == ec2.a == expect_ec.a

    def test_create_from_multiple_specific_class_dicts_different_class(self):
        self.write_example_class_dict_to_file()  # example_class
        self.write_example_class_dict_to_file(2)  # second_example_class
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=DIFFERENT_CLASS_CONFIG_DICT_LIST
        )
        pipeline_manager.load()
        sel = Selector()
        ec = sel.test_pipeline_manager.example_class.stuff.data
        sec = sel.test_pipeline_manager.second_example_class.stuff.data
        expect_ec = ExampleClass(None, name='data')
        expect_sec = SecondExampleClass(name='data')
        assert ec.name == expect_sec.name == expect_ec.name
        assert ec.a == expect_ec.a
        assert sec.b == expect_sec.b



"""
Flow Manager is a Python framework for flow-based programming and managing configuration. To use it, you register
your main functions, and config files are created automatically for them. It provides a way to run these functions
individually, in a list, in sections, or a combination thereof. Configuration has multiple layers, with the defaults
for the function being the lowest level config, which then gets overrided by project config, which gets overrided by
the config file, and finally by any local config updates in a script.
"""

from pyfileconf.main import PipelineManager
from pyfileconf.selector.models.selector import Selector



from typing import Dict, Type, List, Any

from pyfileconf.main import PipelineManager


def get_dict_of_pipeline_manager_names_and_instances_from_globals(globs=None) -> Dict[str, PipelineManager]:
    managers = _get_objs_of_class_from_globals(PipelineManager, globs=globs)
    out_dict: Dict[str, PipelineManager] = {}
    for manager in managers:
        name = manager.name
        if name in out_dict:
            # Two cases where name may already be in dict. User has defined two pipeline managers with the same name,
            # or the same pipeline manager has been assigned to another variable.

            # Here the same pipeline manager has been assigned to multiple variables. Just skip adding this time
            if manager is out_dict[name]:
                continue

            # Here this is actually a different pipeline manager than the one stored. They both have the same name,
            # so this must be a user error of defining multiple pipeline managers without different names
            raise ValueError(
                f'found multiple PipelineManager objects with the name {name}, cannot determine which to select')
        out_dict[name] = manager

    return out_dict


def _get_objs_of_class_from_globals(klass: Type, globs=None) -> List:
    if globs is None:
        globs = globals()

    selected_items = []
    for variable_name, value in globs.items():
        if isinstance(value, klass):
            selected_items.append(value)

    return selected_items


def _get_dicts_of_variable_name_value_from_globals_by_class(klass: Type, globs=None) -> Dict[str, Any]:
    if globs is None:
        globs = globals()

    selected_dict: Dict[str, Any] = {}
    for variable_name, value in globs.items():
        if isinstance(value, klass):
            if variable_name in selected_dict:
                raise ValueError(
                    f'found multiple {klass} objects assigned to {variable_name}, cannot determine which to select')
            selected_dict[variable_name] = value

    return selected_dict


from typing import List
from pyfileconf.sectionpath.sectionpath import SectionPath

accepted_name_attrs = ['name', '__name__']
output_accepted_name_attrs = ['output_name'] + accepted_name_attrs

def _get_from_nested_obj_by_section_path(obj, section_path: SectionPath):
    # Goes into nested sections, until it pulls the final section or pipeline
    for section in section_path:
        obj = getattr(obj, section)

    return obj


def _get_public_name_or_special_name(obj, accept_output_names=True) -> str:
    name_attrs = _get_name_attrs(accept_output=accept_output_names)
    for name_var in name_attrs:
        if hasattr(obj, name_var):
            return getattr(obj, name_var)

    raise ValueError(f'could not get .name or .__name__ from {obj} of type {type(obj)}')

def _has_public_name_or_special_name(obj, accept_output_names=True) -> bool:
    name_attrs = _get_name_attrs(accept_output=accept_output_names)
    for name_var in name_attrs:
        if hasattr(obj, name_var):
            return True

    return False

def _get_name_attrs(accept_output: bool=True) -> List[str]:
    if accept_output:
        return output_accepted_name_attrs
    else:
        return accepted_name_attrs


import builtins
from typing import Any


def _is_builtin(value: Any) -> bool:
    if value is None:
        return True # None won't return True from the following check

    builtin_types = [getattr(builtins, d) for d in dir(builtins) if isinstance(getattr(builtins, d), type)]
    return type(value) in builtin_types

def _is_str_matching_builtin_type(str_value: str) -> bool:
    """
    should pass 'str', 'int', etc.
    """
    builtin_types = [getattr(builtins, d) for d in dir(builtins) if isinstance(getattr(builtins, d), type)]
    return f"<class '{str_value}'>" in [str(bt) for bt in builtin_types]


import inspect


def iter_globals():
    """
    Generator which first returns globals for current module, then next level up, and so on
    """
    for frame in _iter_frames():
        yield frame.f_globals
    for frame in _iter_frames():
        yield frame.f_locals

def _iter_frames():
    """
    Generator which first returns current code frame, then next level up, and so on
    """
    frame = inspect.currentframe()
    while frame is not None:
        yield frame
        frame = frame.f_back  # get next level up frame






from pyfileconf.io.file.load.lazy.pipeline import PipelineDictLoader

class PipelineDictInterface(PipelineDictLoader):

    def save(self):
        # TODO [#9]: write pipeline dict to file
        pass



from pyfileconf.io.file.load.lazy.datadict import SpecificClassDictLoader

class SpecificClassInterface(SpecificClassDictLoader):

    def save(self):
        # TODO [#8]: write data dict to file
        pass